;Erweiterung der Ausgabe-Funktionalitt fr LCD fr AVR-Assembler (AT-Mega).
;Dieser Include-Code ist Bestandteil LCD-Ausgaberoutinensammlung "print.inc".

;Die Datenausgaberoutine des LCD-Treibers sollte "lcd_data" heien.
;Das auszugebende ASCII-Zeichen steht im Register "tmp", das im Hauptprogramm
;eingerichtet werden muss.

;Verwendete Register und SRAM:
;  Reg.: tmp (wird nicht gesichert)
;        wh, zh und zl (wird gesichert)
;  SRAM: tmpmath(0,1)
;-----------------------------------------------------------------------------

;Es werden folgende Funktionen bereitgestellt,
;die per Macro aufgerufen werden:
;-----------------------------------------------------------------------------
;  print16 H-reg,L-Reg,Komma,[using]+[sig]+[fnul]
                        ;Gibt 16-Bit-Zahl als Ziffern aus
                        ;H-Reg,L-Reg enthlt die Zahl
                        ;Komma ist die Anzahl der Nachkommastellen
                        ;using erzwingt formatierte Ausgabe (Print Using)
                        ;sig erzwingt Vorzeichen
                        ;fnul erzwingt Fhrungsnullen
;-----------------------------------------------------------------------------
;  print16s Adr,Komma,[using]+[sig]+[fnul]
                        ;Gibt 16-Bit-Zahl in SRAM als Ziffern aus
                        ;Adr und ADR+1 (SRAM) enthlt die Zahl (L:H)
                        ;Komma ist die Anzahl der Nachkommastellen
                        ;using erzwingt formatierte Ausgabe (Print Using)
                        ;sig erzwingt Vorzeichen
                        ;fnul erzwingt Fhrungsnullen
;-----------------------------------------------------------------------------
;  printui H-Reg,L-Reg  ;Gibt den Inhalt der Zahl in H-reg:L-Reg 5-stellig aus
                        ;Ist inzwischen dank "print16" berflssig
;-----------------------------------------------------------------------------
.equ using=0b00100000   ;Flag fr formatierte Ausgabe (Print Using)
.equ sig = 0b01000000   ;Flag fr Vorzeichen-Zahl (signed Integer)
.equ fnul= 0b10000000   ;Flag fr Fhrungsnullen
;-----------------------------------------------------------------------------

.macro print16 ;H-Reg,L-Reg,Kommastellen,using+sign+fnul
 sts tmpmath,@1             ;Low-Byte und
 sts tmpmath+1,@0           ;High-Byte ins temp. SRAM (bergabebereich)
 ldi tmp,(@3 & $e0)|16|(6-(@2 & 15)) ;Mode-Flags und Komma-Position direkt
 rcall lcd_print16          ;Aufruf
.endmacro



.macro print16s ;Adresse,Kommastellen,using+sign+fnul
 
 lds tmp,@0                 ;Daten
 sts tmpmath,tmp            ;aus SRAM in
 lds tmp,@0+1               ;bergabebereich
 sts tmpmath+1,tmp          ;kopieren
 ldi tmp,(@2 & $e0)|16|(6-(@1 & 15))   ;Mode-Flags und Komma-Position direkt
 rcall lcd_print16          ;Aufruf
.endmacro

.macro printui ;H-Register, L-Register
                        ;Gibt u-Integer im Registerpaar H:L 5-stellig an LCD aus.
                        ;Die Parameter geben die Register H und L an.
                        ;Diese Register werden nicht verndert.
                        ;Kopie der Zahl in XH und XL
 sts tmpmath,@1             ;Low-Byte und
 sts tmpmath+1,@0           ;High-Byte ins temp. SRAM (bergabebereich)
 ldi tmp,sig+fnul           ;mit Fhrungsnullen ausgeben
 rcall lcd_print16          ;Aufruf
.endmacro

;-----------------------------------------------------------------------------

lcd_print16:            ;Wird von Macro aufgerufen. Gibt Word als Zahl in mehreren Modi aus
 push xh                    ;verwendete Register
 push xl                    ;sichern
 push wh
 mov wh,tmp                 ;Format-Bitmuster kopieren
 lds xl,tmpmath             ;Wert aus bergabe-
 lds xh,tmpmath+1           ;Bereich holen

 sbrs wh,6                  ;Signum-Flag gesetzt? ja...
 rjmp lcd_printui           ;nein, als unsigned-Int behandeln...
 sbrs xh,7                  ;Wert negativ? ja...
 rjmp lcd_printsig          ;nein, positiv...
 ldi tmp,'-'                ;Vorzeichen
 rcall lcd_data             ;ausgeben
 neg xl                     ;Zahl positiv
 com xh                     ;machen
lcd_printui:            ;
 ldi tmp,-1+'0'             ;Ziffer fr Zehntausender
 inc tmp                    ;Zehntausender Stelle hoch und
 subi xl,low(10000)         ;von Zahl 10000 subtrahieren
 sbci xh,high(10000)        ;bis Zahl negativ wird
 brcc pc-3                  ;schon negativ? nein...
lcd_print16_a:          ;Einsprung aus 24-Bit-Routine
 rcall lcd_format           ;ja, Ziffer ausgeben (Zehntausender)
 ldi tmp,10+'0'             ;Ziffer fr Tausender, Zahl ist jetzt negativ
 dec tmp                    ;Tausender Stelle verkleinern und
 subi xl,low(-1000)         ;1000 zur (negativen) Zahl addieren,
 sbci xh,high(-1000)        ;bis Zahl wieder positiv wird
 brcs pc-3                  ;noch negativ? ja...
 rcall lcd_format           ;nein, Ziffer ausgeben (Tausender)
 ldi tmp,-1+'0'             ;Ziffer fr Hunderter (Zahl ist jetzt positiv)
 inc tmp                    ;Hunderter Stelle hoch und
 subi xl,low(100)           ;von Zahl 100 subtrahieren
 sbci xh,high(100)          ;bis Zahl negativ wird
 brcc pc-3                  ;schon negativ? nein...
 rcall lcd_format           ;ja, Ziffer ausgeben (Hunderter)
 ldi tmp,10+'0'             ;Zehner-Stelle als ASCII-Zeichen, Zahl ist negativ
 dec tmp                    ;Zehner runter und
 subi xl,-10                ;10 addieren bis positiv wird
 brlo pc-2                  ;positiv? nein, 2 Zeilen hoch...
 rcall lcd_format           ;ja, Zehner Stelle ausgeben...
 ldi tmp,'0'                ;ASCII-0
 add tmp,xl                 ;Einer addieren (Rest war ja positiv)
 rcall lcd_format           ;Einer ausgeben
 andi wh,using              ;Fhrungsnullen oder Using aktiv?
 brne pc+3                  ;ja, also konstante Textlnge...
 ldi tmp,' '                ;nein, variable Textlnge, deshalb
 rcall lcd_data             ;Leerzeichen anhngen um alte Ziffer zu lschen
 pop wh
 pop xl                     ;verwendete Register
 pop xh                     ;wiederherstellen
 ret                        ;zurck

lcd_printsig:
 ldi tmp,'+'                ;Vorzeichen
 rcall lcd_data             ;ausgeben
 rjmp lcd_printui           ;Zahl ausgeben

lcd_format:
 dec wh                     ;Kommazhler runter
 push wh                    ;Mode-Flags sichern
 andi wh,$0f                ;Kommazhler isolieren
 cpi wh,1                   ;Letzte Stelle vor dem Komma?
 pop wh                     ;(Mode-Flags wegen Vernderung holen)
 brne pc+2                  ;nein...
 sbr wh,fnul                ;Nullausgabe aktivieren
 push wh                    ;genderte Mode-Flags nochmal sichern
 andi wh,$0f                ;Kommazhler isolieren (wegen Zero-Flag)
 pop wh                     ;Mode-Flags wiederherstellen
 brne lcd_keinkomma         ;Kommazhler nicht Null...
 push tmp                   ;Ziffer sichern
 ldi tmp,','                ;Komma
 rcall lcd_data             ;ausgeben
 pop tmp                    ;Ziffer wiederherstellen
lcd_keinkomma:
 cpi tmp,'0'                ;Ziffer = "0"?
 breq lcd_format0           ;ja...
 sbr wh,fnul                ;nein, knftige Nullen nicht mehr unterdrcken
 rcall lcd_data             ;Ziffer ausgeben 
 ret                        ;und zurck...

lcd_format0:
 sbrs wh,7                  ;Nullen ausgeben? ja...
 ldi tmp,' '                ;nein, Null durch Leerzeichen ersetzen
 sbrc wh,7                  ;Nullen ausgeben? nein...
 rcall lcd_data             ;ja, Ziffer "0" ausgeben 
 ret                        ;und zurck...
 sbrc wh,5                  ;Print Using? nein...
 rcall lcd_data             ;ja, Leerzeichen ausgeben und zurck...
 ret                        ;nein, ohne Ausgabe zurck...

